// Quad_ROV_Controller_ESC_PID_Echo_Serial_3
// Written by Theron WIerenga, June 2017
// Changes September 18, 2017

/*
  Quad motor configuration

   ->     <- 
    M0   M1
      \ /
      / \
    M3   M2
   ->     <-

*/

#define PRESSURE A0     // Port to read pressure, no reduction to 3.3 volts
#define BUFFER_SIZE 30  // Average 30 reads for sensors, helps reduce errors from vibration
#define MULTIPLEX 2     // For switching servo signal into ESCs
#define FLAG 9          // For blue LED
#define ARMED A2        // Denotes ESCs are receiving servo signals from 
                        // Adafruit Servo Driver

// Yhese values may need to be changed depending on ESCs and motors used
// MIN_SIGNAL must ensure that motor is always running, ie. just running
#define MAX_SIGNAL 360  // = 1911 uSec
#define MIN_SIGNAL 212  // = 1127 uSec

// Include needed libraries
#include <PID_v1.h>
#include <Wire.h>
#include <Adafruit_Sensor.h>
#include <Adafruit_LSM303_U.h>
#include <Adafruit_L3GD20_U.h>
#include <Adafruit_9DOF.h>
#include <Adafruit_PWMServoDriver.h>

/* Assign a unique ID to the sensors */
Adafruit_9DOF                 dof   = Adafruit_9DOF();
Adafruit_LSM303_Accel_Unified accel = Adafruit_LSM303_Accel_Unified(30301);
Adafruit_LSM303_Mag_Unified   mag   = Adafruit_LSM303_Mag_Unified(30302);
Adafruit_PWMServoDriver pwm = Adafruit_PWMServoDriver();

/* Update this with the correct SLP for accurate altitude measurements */
float seaLevelPressure = SENSORS_PRESSURE_SEALEVELHPA;
sensors_event_t accel_event;
sensors_event_t mag_event;
sensors_vec_t   orientation;

String inputString = "", strTemp = "", Gstr = "", Tstr = "";
boolean stringComplete = false, debug = false;

double range, min_double, max_double, divisor = 270.0;

//Define Variables we'll be connecting to
double rollSetpoint, rollInput, rollOutput;
double pitchSetpoint, pitchInput, pitchOutput;
double yawSetpoint, yawInput, yawOutput;
double depthSetpoint, depthInput, depthOutput;
int temp, depthInches, dcount = 0;
int depthindexBuffer = 0, rollindexBuffer = 0, pitchindexBuffer = 0, yawindexBuffer = 0;
float depthSum, depthBuffer[BUFFER_SIZE], rollSum, rollBuffer[BUFFER_SIZE], pitchSum, pitchBuffer[BUFFER_SIZE], yawSum, yawBuffer[BUFFER_SIZE];
int motors[4];

// Define the aggressive and conservative Tuning Parameters
// double aggKp = 4, aggKi = 0.2, aggKd = 1;
// double midKp = 2.5, midKi = 0.125, midKd = 0.625;
// double consKp = 1, consKi = 0.05, consKd = 0.25;
// Penn State quad project had Kp = 2.0, Kd = 0.6 to start
//                             Kp = 1.5, Kd = 0.3 at end
double dKp = 3.0, dKi = 0.0, dKd = 0.3;
double rKp = 2.5, rKi = 0.0, rKd = 0.5;
double pKp = 2.5, pKi = 0.0, pKd = 0.5;
double yKp = 2.5, yKi = 0.0, yKd = 0.5;
 
//Specify the links and initial tuning parameters
PID rollPID(&rollInput, &rollOutput, &rollSetpoint, rKp, rKi, rKd, DIRECT);
PID pitchPID(&pitchInput, &pitchOutput, &pitchSetpoint, pKp, pKi, pKd, DIRECT);
PID yawPID(&yawInput, &yawOutput, &yawSetpoint, yKp, yKi, yKd, DIRECT);
PID depthPID(&depthInput, &depthOutput, &depthSetpoint, dKp, dKi, dKd, REVERSE);

void setup()
{
    // Allow ESCs to initialize
    pinMode(MULTIPLEX, OUTPUT);
    digitalWrite(MULTIPLEX, HIGH); // Make sure 555 timer is going to ESCs
    
    pinMode(FLAG, OUTPUT);          // Blue LED
    digitalWrite(FLAG, LOW);
  
    // Setup PWM
    pwm.begin();
    pwm.setPWMFreq(50);                 // Analog servos run at ~50 Hz updates
    pwm.setPWM(0, 0, MIN_SIGNAL - 5);   // Turn them all off
    pwm.setPWM(1, 0, MIN_SIGNAL - 5);
    pwm.setPWM(2, 0, MIN_SIGNAL - 5);
    pwm.setPWM(3, 0, MIN_SIGNAL - 5);

    min_double = (double)MIN_SIGNAL;    // Convert to double floating point
    max_double = (double)MAX_SIGNAL;
    range = max_double - min_double;

     // Initialise the sensors
    initSensors();

    // Smooth out initial sensor readings
    for (int i = 0; i < BUFFER_SIZE; i++)
    {
       temp = analogRead(PRESSURE);
       readSensors();
       depthInput = smoothdepthSensorReadings(temp);
       rollInput = smoothrollSensorReadings(orientation.roll);
       pitchInput = smoothpitchSensorReadings(orientation.pitch);
       yawInput = smoothyawSensorReadings(orientation.heading);
    }

    // Initialize PID variables
    // Inputs were set in the initialize loop above
    rollOutput = 0;
    pitchOutput = 0;
    yawOutput = 0;
    depthOutput = 0;
    rollSetpoint = 0;         // these change with joysticks
    pitchSetpoint = 0;
    yawSetpoint =  yawInput;
    depthSetpoint = 528;
   
    // Turn the PIDs on
    rollPID.SetMode(AUTOMATIC);
    rollPID.SetSampleTime(10);
    rollPID.SetOutputLimits(-255, 255);
    pitchPID.SetMode(AUTOMATIC);
    pitchPID.SetSampleTime(10);
    pitchPID.SetOutputLimits(-255, 255);
    yawPID.SetMode(AUTOMATIC);
    yawPID.SetSampleTime(10);
    yawPID.SetOutputLimits(-255, 255);
    depthPID.SetMode(AUTOMATIC);
    depthPID.SetSampleTime(10);
    depthPID.SetOutputLimits(-255, 255);
  
    Serial3.begin(19200);         // RX, TX  now (7, 8) for Teensy 3.1
    // reserve 300 bytes for the inputString:
    inputString.reserve(300);
    delay(8000);                  // Wait for ESCs to initialize
    digitalWrite(MULTIPLEX, LOW); // ESCs should be powered up so switch input
    pinMode(ARMED, OUTPUT);
    digitalWrite(ARMED, HIGH);    // Turn on Teensy LED to show ESCs are armed
    Serial.begin(115200);
    Serial.println("Starting"); 
    Serial.println();
    digitalWrite(FLAG, HIGH);     // Flash blue LED
    delay(1000);
    digitalWrite(FLAG, LOW);

    // Send dummy joystick movements to be sure communication to joystick controller is working
    Serial.println("Sending"); 
    Serial3.println("  ");
    delay(100);
    Serial3.println("R0");
    delay(100);
    Serial3.println("P0");
    delay(100);
    Serial3.println("Y0");
    delay(100);
    Serial3.println("D0");
    delay(100);
    Serial3.println("G0");
    delay(100);
    Serial3.println("T0");
    delay(100);
    Serial3.println("A0");
    delay(100);
    Serial.println("End Sending"); 
}

void loop()
{
    mySerialEvent();      // Has serial data been received from joystick controller
    if (stringComplete)   // There is new data
    {
        Serial3.println(inputString);                   // Echo command back to the joystick controller
        Serial.println(inputString);
        String subStr = inputString.substring(0, 1);    // Get the letter
        int len = inputString.length() - 1;
        String temp = inputString.substring(1, len);
        float val = temp.toFloat();                     // Get the value
        inputString = "";                               // Clear for next data read
        stringComplete = false;
        
        if (subStr == "R") { rollSetpoint = val * 2.0; }    // Change roll, pitch, yaw, or depth setpoint
        if (subStr == "P") { pitchSetpoint = val * 2.0; }
        if (subStr == "Y") 
        { 
          yawSetpoint = yawSetpoint + val; 
          if (yawSetpoint > 360.0) { yawSetpoint -= 360.0; }
          if (yawSetpoint < 0.0) { yawSetpoint += 360.0; }
        }
        
        if (subStr == "D") 
        { 
          depthSetpoint = depthSetpoint + val; 
          if (depthSetpoint < 377) { depthSetpoint = 377; }
          if (depthSetpoint > 766) { depthSetpoint = 766; }  // 50 feet
        }
// This section reads the joystick switches and was used during development
// It can be used to test different parameters, like the PID gain constants
// min_double, max_double, divisor, tec. Changes are display on the joystick
// controller LCD

         if (subStr == "S") 
        { 
          if (val > 0)                                // Left switch
          {
//            rKp = rKp + 0.25;
//            pKp = pKp + 0.25;
//            if (rKp > 5.0) { rKp = 1.0; }
//            if (pKp > 5.0) { pKp = 1.0; }
//            // depthPID.SetTunings(dKp, dKi, dKd);
//            rollPID.SetTunings(rKp, rKi, rKd);
//            pitchPID.SetTunings(pKp, pKi, pKd);
//            Serial3.print("S");
//            Serial3.println(rKp);  
              min_double = min_double + 1.0;
              Serial3.print("S");
              Serial3.println(min_double);  
          }
          else                                        // Right switch
          {
//            rKd = rKd + 0.1;
//            pKd = pKd + 0.1;
//            if (rKd > 3.00) { rKd = 0.1; }
//            if (pKd > 3.00) { pKd = 0.1; }
//            // depthPID.SetTunings(dKp, dKi, dKd);
//            rollPID.SetTunings(rKp, rKi, rKd);
//            pitchPID.SetTunings(pKp, pKi, pKd);
              divisor = divisor + 2.0;
              Serial3.print("S");
              Serial3.println(divisor);  
          }
            
        }
    }

    dcount += 1;
    if (dcount == 300)    // We need to send the depth to the joystick controller periodically
    {
        depthInches = int(((depthInput - 518.0) * 1.05) + 0.5); 
        strTemp = "A" + String(depthInches);                    // Actual depth in inches
        Serial3.println(strTemp);
        Gstr = "G" + String(int(depthInput));                   // Depth input from pressure guage
     //   Serial.println(Gstr);
        Serial3.println(Gstr);
        Tstr = "T" + String(int(depthSetpoint));                // Depth setpoint - target
      //  Serial.println(Tstr);
        Serial3.println(Tstr);
        dcount = 0;
    }

    readSensors();
    temp = analogRead(PRESSURE);                                      // Read the sensors
    depthInput = double(smoothdepthSensorReadings(temp));
    rollInput = smoothrollSensorReadings(orientation.roll);
    pitchInput = smoothpitchSensorReadings(orientation.pitch);
    yawInput = smoothyawSensorReadings(orientation.heading);
    depthPID.Compute();
    rollPID.Compute();                                          // PIDs will conpute the new output values
    pitchPID.Compute();
//    yawPID.Compute(); // Not implemented - will need further tuning

//    Serial.print(rollInput);
//    Serial.print("    ");
//    Serial.println(pitchInput);

    // Add the outputs for the four motors and scale the value with the divisor variable
    motors[0] = (int)(((depthOutput + rollOutput + pitchOutput + yawOutput) / divisor) * range + min_double);
    motors[1] = (int)(((depthOutput - rollOutput + pitchOutput - yawOutput) / divisor) * range + min_double);
    motors[2] = (int)(((depthOutput - rollOutput - pitchOutput + yawOutput) / divisor) * range + min_double);
    motors[3] = (int)(((depthOutput + rollOutput - pitchOutput - yawOutput) / divisor) * range + min_double);

    // Clip the motor values if necessary
    if (motors[0] > (int)max_double) { motors[0] = (int)max_double; }
    if (motors[1] > (int)max_double) { motors[1] = (int)max_double; }
    if (motors[2] > (int)max_double) { motors[2] = (int)max_double; }
    if (motors[3] > (int)max_double) { motors[3] = (int)max_double; }
    
    if (motors[0] < (int)min_double) { motors[0] = (int)min_double; }
    if (motors[1] < (int)min_double) { motors[1] = (int)min_double; }
    if (motors[2] < (int)min_double) { motors[2] = (int)min_double; }
    if (motors[3] < (int)min_double) { motors[3] = (int)min_double; }

    // Send the motor values to the Adafruit Servo Driver board
    // This sets the servo signal that drives the ESCs
    pwm.setPWM(0, 0, motors[0]);
    pwm.setPWM(1, 0, motors[1]);
    pwm.setPWM(2, 0, motors[2]);
    pwm.setPWM(3, 0, motors[3]);
   // Serial.println(millis());
}

// Reads incoming serial from joystick controller places it in inputString
void mySerialEvent() 
{
    while ((Serial3.available() > 0) && (stringComplete == false))
    {
        digitalWrite(FLAG, HIGH);
        // get the new byte:
        char inChar = (char)Serial3.read();
        // add it to the inputString:
        inputString += inChar;
       // delay(2);
        // if the incoming character is a newline, set a flag
        // so the main loop can do something about it:
        if (inChar == '\n') 
        {
          stringComplete = true;
        }
    }
    digitalWrite(FLAG, LOW);
}

// Initializes the sensors
void initSensors()
{
  if(!accel.begin())
  {
    /* There was a problem detecting the LSM303 ... check your connections */
    Serial.println(F("Ooops, no LSM303 detected ... Check your wiring!"));
    while(1);
  }
  if(!mag.begin())
  {
    /* There was a problem detecting the LSM303 ... check your connections */
    Serial.println("Ooops, no LSM303 detected ... Check your wiring!");
    while(1);
  }
}

// Reads the sensors
void readSensors()
{
        // Calculate pitch and roll from the raw accelerometer data
    accel.getEvent(&accel_event);
    if (dof.accelGetOrientation(&accel_event, &orientation))
    {
//      if (debug)
//      {
//        // 'orientation' should have valid .roll and .pitch fields
//        Serial.print(F("Roll: "));
//        Serial.print(orientation.roll);
//        Serial.print(F("; "));
//        Serial.print(F("Pitch: "));
//        Serial.print(orientation.pitch);
//        Serial.print(F("; "));
//      }
    }
    
    // Calculate the heading using the magnetometer
    mag.getEvent(&mag_event);
    if (dof.magGetOrientation(SENSOR_AXIS_Z, &mag_event, &orientation))
    {
//      if (debug)
//      {
//        // 'orientation' should have valid .heading data now
//        Serial.print(F("Heading: "));
//        Serial.print(orientation.heading);
//        Serial.println(F("; "));
//      }
    }
}

// The next four methods uses circular buffers to smooth out the sensor readings
float smoothdepthSensorReadings(float input)
{
  // We remove the oldest value from the buffer
    depthSum = depthSum - depthBuffer[depthindexBuffer];
  // The new input from the sensor is placed in the buffer
    depthBuffer[depthindexBuffer] = input;
  // It is also added to the total sum of the buffer
  // This method avoids needing to sum all the elements every time this function is called.
    depthSum += input;
  // We increment the pointer
    depthindexBuffer++;
  // We test if we arrived at the end of the buffer, in which case we start again from index 0
    if (depthindexBuffer == BUFFER_SIZE) { depthindexBuffer = 0; }
  // The output is the mean value of the circular buffer.
    return (depthSum/BUFFER_SIZE);
}

float smoothrollSensorReadings(float input)
{
    rollSum = rollSum - rollBuffer[rollindexBuffer];
    rollBuffer[rollindexBuffer] = input;
    rollSum += input;
    rollindexBuffer++;
    if (rollindexBuffer == BUFFER_SIZE) { rollindexBuffer = 0; }
    return (rollSum/BUFFER_SIZE);
}

float smoothpitchSensorReadings(float input)
{
    pitchSum = pitchSum - pitchBuffer[pitchindexBuffer];
    pitchBuffer[pitchindexBuffer] = input;
    pitchSum += input;
    pitchindexBuffer++;
    if (pitchindexBuffer == BUFFER_SIZE) { pitchindexBuffer = 0; }
    return (pitchSum/BUFFER_SIZE);
}

float smoothyawSensorReadings(float input)
{
    yawSum = yawSum - yawBuffer[yawindexBuffer];
    yawBuffer[yawindexBuffer] = input;
    yawSum += input;
    yawindexBuffer++;
    if (yawindexBuffer == BUFFER_SIZE) { yawindexBuffer = 0; }
    return (yawSum/BUFFER_SIZE);
}

